//:Shape.java -- defines Shape with Polyhedron
//interface and subclasses thereof
package com.pdx4d.polys;

interface Polyhedron {
    void rotate(String axis, double degrees);
    void translate(double[] vector);
    void scale(double scalefactor);
}

public abstract class Shape implements Polyhedron {

    // Math constants used in constructor volume & scaling calcs
    public static final double SYN3 = Math.sqrt(9.0/8.0);
    public static final double ROOT2= Math.sqrt(2.0);
    public static final double ROOT5= Math.sqrt(5.0);
    public static final double PHI  = (1.0 + ROOT5)/2;

    // properties of Shapes (not all subclasses use them all)

    public String
        edgecolor,
        vertcolor,
        shapefiletype="Q",
        keepfiletype;

    public boolean
        edgesvisible = true,
        meshable = false,
        meshed = false,
        vrmlmesh = false,
        colorcoded = false,
        curved = false,
        mute = true;

    public int
        noverts=0,
        noedges=0,
        frequency=0;    // mesh frequency if meshable

    public int[][]
         edges,
         faces;  // not used by all polys -- for applying meshes

    public double
        cylradius = 0.04,
        vertradius = 0.04,
        pointradius = 0,
        edgefactor = 1,
        shapevolume = 0,
        radius = 0;

    public double[]
        pivot = new double[3],
        lastpos = new double[3],
        orient = new double[3];

    public double[][]
        points,
        rgbcodes; // used only for color coded nubs (specialized)

    public MeshManager oMesh = null;
    public Mesh[] meshes;  // array of child meshes (specific kind of shape)

    // Matrix object used internally for calcs on points and
    // holds some data potentially unique to each instantiation
    protected Matrix oMatrix;

    // The Formats class gets used directly without any objects
    // of that type being instantiated (is an abstract class
    // with static methods)

    // When initialized, pass a reference to this object to
    // the internalized Matrix object, so it will be able to
    // locate this shape's point set (if relevant).

    Shape(){
        oMatrix = new Matrix(this);
    }

    public void setcolor(String shapecolor){
        edgecolor = shapecolor;
        vertcolor = shapecolor;
        if (meshed) oMesh.setmeshcolor(shapecolor, shapecolor);
    }

    public void setcolor(String ecolor, String vcolor){
        edgecolor = ecolor;
        vertcolor = vcolor;
        if (meshed) oMesh.setmeshcolor(ecolor, vcolor);
    }

    public void setdim(double e, double v){
        cylradius = e;
        vertradius = v;
        if (meshed) oMesh.setmeshdim(e,v);
    }

    public void rotate(String axis, double degrees){
       oMatrix.setdegrees(axis, degrees);
       rotate();
    }

    // uses previous axis, degrees stored in oMatrix object
    public void rotate(){
       convert2xyz();
       if (!meshed){
            gohome();
            oMatrix.rotate();
            golast();
       } else if (!curved){
            meshed = false;
            gohome();
            oMatrix.rotate();
            golast();
            meshed = true;
            oMesh.remesh();
       } else {
            meshed = false;
            gohome();
            oMatrix.rotate();
            meshed = true;
            oMesh.remesh();
            oMesh.curvemesh();
            golast();
       }
    }

  public void reorient(double[] axes){
      rotate("X",axes[0]);
      rotate("Y",axes[1]);
      rotate("Z",axes[2]);
  }

  public void alignwith(Shape s){
      double[] axes = new double[3];
      for (int i=0;i<3;i++) axes[i]=s.orient[i]-orient[i];
      reorient(axes);
  }

    public void translate(double[] d){
        convert2xyz();
        if (d.length == 4) d = Formats.quad2xyz(d);
        oMatrix.translate(d);
        for (int i=0;i<3;i++) lastpos[i]=pivot[i];
        for (int i=0;i<3;i++) pivot[i]=pivot[i]+d[i];
        if (meshed) {
        if (!curved) oMesh.remesh();
              else oMesh.translate(d);
            }
    }

    public void gohome(){
        if (pivot[0]!=0.0 | pivot[1]!=0.0 | pivot[2]!=0.0)
            translate(new double[]{-pivot[0],-pivot[1],-pivot[2]});
    }

    public void golast(){
        translate(new double[]{lastpos[0],lastpos[1],lastpos[2]});
    }

    public void scale(double scalefactor){
        if (scalefactor==0) return;
        convert2xyz();
        oMatrix.scale(scalefactor);
        shapevolume = shapevolume * Math.pow(scalefactor,3);
        edgefactor = edgefactor * scalefactor;
        if (meshed) oMesh.remesh();
    }

    public void scalewith(Shape s){
        double scalefactor = s.edgefactor/edgefactor;
        scale(scalefactor);
    }

    public void synchwith(Shape s){
        scalewith(s);
        alignwith(s);
        translate(new double[]{s.pivot[0]-pivot[0],s.pivot[1]-pivot[1],s.pivot[2]-pivot[2]});
    }

// Point conversion public methods

    public void convert2xyz(){
        if (!shapefiletype.equals("X")){
              keepfiletype=shapefiletype;
              if (shapefiletype.equals("P")) points=Formats.polar2xyz(points);
              if (shapefiletype.equals("Q")) points=Formats.quad2xyz(points);
              shapefiletype="X";
        }
        if (meshed) oMesh.convert2xyz();
    }

    public void convert2quad(){
        if (!shapefiletype.equals("Q")){
              keepfiletype=shapefiletype;
              if (shapefiletype.equals("X")) points=Formats.xyz2quad(points);
              if (shapefiletype.equals("P")) points=Formats.polar2quad(points);
              shapefiletype="Q";
        }
        if (meshed) oMesh.convert2quad();
    }

    public void convert2polar(){
        if (!shapefiletype.equals("P")){
            keepfiletype=shapefiletype;
            if (shapefiletype.equals("X")) points=Formats.xyz2polar(points);
            if (shapefiletype.equals("Q")) points=Formats.quad2polar(points);
            shapefiletype="P";
        }
        if (meshed) oMesh.convert2polar();
    }

    public void convert2last(){
        if (keepfiletype.equals("Q")) convert2quad();
        if (keepfiletype.equals("X")) convert2xyz();
        if (keepfiletype.equals("P")) convert2polar();
    }

// Mesh-related public methods

    public void addmesh(int f){
       if (!meshable) return;
       convert2xyz();
       oMesh = new MeshManager(f);
       meshed = true;
       frequency = f;
    }

    public void removemesh(int f){
        oMesh = null;
        meshed = false;
        curved = false;
        frequency = 0;
    }

    public void curvemesh(){
        if (meshed & !curved){
          gohome();
          oMesh.curvemesh();
          curved=true;
          golast();
        }
    }

    public void uncurvemesh(){
        if (meshed & curved){
            curved=false;
            oMesh.remesh();
        }
    }

    public void setfreq(int f){
        if (meshed & f!=frequency){
            frequency = f;
            oMesh.setfreq(f);
            if (curved) curvemesh();
        }
    }

    public class MeshManager {

     MeshManager(int f){
        Shape targ = Shape.this;
        meshes = new Mesh[targ.faces.length];

        for(int i=0;i<targ.faces.length;i++){
              if (targ.faces[i].length==3) meshes[i] = new Trimesh(targ,targ.faces[i],f);
              if (targ.faces[i].length==4) meshes[i] = new Sqrmesh(targ,targ.faces[i],f);
              meshes[i].vertradius = targ.vertradius;
              meshes[i].cylradius = targ.cylradius;
              meshes[i].setcolor(targ.edgecolor, targ.vertcolor);
        }

        String shapetype = (Shape.this.getClass()).getName();
              if (!Shape.this.mute) System.out.println(shapetype + " meshed");
      }

      public void setmeshcolor(String ecolor, String vcolor){
        for(int i=0;i<meshes.length;i++) meshes[i].setcolor(ecolor, vcolor);
      }

      public void setmeshdim(double e, double v){
        for(int i=0;i<meshes.length;i++) meshes[i].setdim(e, v);
      }

      public void convert2xyz(){
        for(int i=0;i<meshes.length;i++) meshes[i].convert2xyz();
      }

      public void convert2quad(){
          for(int i=0;i<meshes.length;i++) meshes[i].convert2quad();
      }

      public void convert2polar(){
          for(int i=0;i<meshes.length;i++) meshes[i].convert2polar();
      }

      public void convert2last(){
          for(int i=0;i<meshes.length;i++) meshes[i].convert2last();
      }

      public void meshvisible(boolean arg){
          for(int i=0;i<meshes.length;i++) meshes[i].edgesvisible = arg;
      }

      public void remesh(){
          for (int i=0;i<meshes.length;i++){
              meshes[i].setcorners(Shape.this,Shape.this.faces[i]);
              meshes[i].calcpoints();
          }
      }

      public void setfreq(int f){
          for (int i=0;i<meshes.length;i++) meshes[i].setfreq(f);
          Shape.this.frequency = f;
      }

      protected void translate(double[] d){
          for(int i=0;i<meshes.length;i++) meshes[i].translate(d);
      }

      public void curvemesh(){
          for(int i=0;i<meshes.length;i++) meshes[i].curve();
      }

      public void cymkmesh(){
          for (int i=0;i<meshes.length;i++) meshes[i].cymk();
          String shapetype = (Shape.this.getClass()).getName();
          if (!Shape.this.mute) System.out.println(shapetype + " meshed");
      }
   }

   public static void main(String[] args){}
}

